/*
 * Secret of Evermore Sanity Fixer
 * Copyright (C) 2006,2008 emuWorks
 * http://games.technoplaza.net/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <stdio.h>
#include <string.h>

// define this at compile time on big endian systems (i.e. PPC Mac's)
#ifdef BYTE_ORDER_BIGENDIAN
    #define SWAP_ON_BIGENDIAN(w) ((w << 8) | (w >> 8))
#else
    #define SWAP_ON_BIGENDIAN(w) (w)
#endif

struct options {
    int game, version;
    const char *filename;
};

static void fix_game(unsigned char *sram, int game, unsigned long checksum) {
    unsigned short *offset = (unsigned short *)(sram + 2 + (game * 0x331));
    unsigned char temp = checksum + sram[4 + (game * 0x331)];
    int i;
    
    for (i = 3; i < 0x331; ++i) {
        checksum &= 0xFF00;
        checksum |= temp;
        checksum <<= 1;
        
        if (checksum > 0xFFFF) checksum -= 0xFFFF;
        temp = checksum + sram[2 + i + (game * 0x331)];
    }
    
    *offset = SWAP_ON_BIGENDIAN(checksum);
}

static void parse_command_line(int argc, char **argv, struct options *opts) {
    int i;
    
    opts->filename = argv[1];
    opts->game = 1;
    opts->version = 0;
    
    for (i = 1; i < argc; ++i) {
        if (strcmp(argv[i], "--game") == 0) {
            if (++i < argc) {
                switch (*argv[i]) {
                    case '2': opts->game = 2; break;
                    case '3': opts->game = 3; break;
                    case '4': opts->game = 4; break;
                    default: opts->game = 1;
                }
            }
        } else if (strcmp(argv[i], "--version") == 0) {
            if (++i < argc) {
                if (strcmp(argv[i], "europe") == 0) opts->version = 1;
                else opts->version = 0;
            }
        } else {
            opts->filename = argv[i];
        }
    }
}

int main(int argc, char **argv) {
    FILE *f;
    struct options opts;
    char sram[0x2000];
    
    if (argc < 3) {
        fprintf(stderr, "syntax: soe-sanity sram-file.srm "
                        "[ --game 1-4 ] [ --version us | europe ]\n");
        return -1;
    }
    
    parse_command_line(argc, argv, &opts);
    
    if ((f = fopen(opts.filename, "rb")) == NULL) {
        fprintf(stderr, "error: unable to open SRAM file '%s'\n", argv[1]);
        return -1;
    }
    
    fseek(f, 0, SEEK_END);
    
    if (ftell(f) != 0x2000) {
        fprintf(stderr, "error: invalid SRAM file size.\n");
        return -1;
    }
    
    fseek(f, 0, SEEK_SET);
    
    if (fread(sram, 0x2000, 1, f) != 1) {
        fprintf(stderr, "error: unable to read SRAM file.\n");
        return -1;
    }
    
    fclose(f);
    
    fix_game(sram, (opts.game - 1), ((opts.version == 0) ? 0x43F : 0x16FF));
    
    if ((f = fopen(opts.filename, "wb")) == NULL) {
        fprintf(stderr, "error: unable to open SRAM file '%s'\n", argv[1]);
        return -1;
    }
    
    if (fwrite(sram, 0x2000, 1, f) != 1) {
        fprintf(stderr, "error: unable to write SRAM data to file.\n");
        return -1;
    }
    
    fclose(f);
    
    printf("fixed checksum for game %d in file %s (%s version)\n",
        opts.game, opts.filename, ((opts.version == 0) ? "us" : "european"));
    
    return 0;
}

